@@ -1,8 +1,7 @@
 #
 #   irb.rb - irb main module
 #   	$Release Version: 0.9.5 $
-#   	$Revision: 1.7.2.2 $
-#   	$Date: 2005/10/25 06:38:25 $
+#   	$Revision: 18465 $
 #   	by Keiju ISHITSUKA(keiju@ruby-lang.org)
 #
 # --
@@ -23,7 +22,7 @@ require "irb/locale"
 STDOUT.sync = true
 
 module IRB
-  @RCS_ID='-$Id: irb.rb,v 1.7.2.2 2005/10/25 06:38:25 matz Exp $-'
+  @RCS_ID='-$Id: irb.rb 18465 2008-08-10 00:35:14Z nobu $-'
 
   class Abort < Exception;end
 
@@ -85,7 +84,7 @@ module IRB
   end
 
   #
-  # irb interpriter main routine 
+  # irb interpreter main routine 
   #
   class Irb
     def initialize(workspace = nil, input_method = nil, output_method = nil)
@@ -149,19 +148,25 @@ module IRB
             line.untaint
 	    @context.evaluate(line, line_no)
 	    output_value if @context.echo?
-	  rescue StandardError, ScriptError, Abort
-	    $! = RuntimeError.new("unknown exception raised") unless $!
-	    print $!.class, ": ", $!, "\n"
-	    if  $@[0] =~ /irb(2)?(\/.*|-.*|\.rb)?:/ && $!.class.to_s !~ /^IRB/
+	    exc = nil
+	  rescue Interrupt => exc
+	  rescue SystemExit, SignalException
+	    raise
+	  rescue Exception => exc
+	  end
+	  if exc
+	    print exc.class, ": ", exc, "\n"
+	    if exc.backtrace[0] =~ /irb(2)?(\/.*|-.*|\.rb)?:/ && exc.class.to_s !~ /^IRB/ &&
+                !(SyntaxError === exc)
 	      irb_bug = true 
 	    else
 	      irb_bug = false
 	    end
-	    
+
 	    messages = []
 	    lasts = []
 	    levels = 0
-	    for m in $@
+	    for m in exc.backtrace
 	      m = @context.workspace.filter_backtrace(m) unless irb_bug
 	      if m
 		if messages.size < @context.back_trace_limit
@@ -183,8 +188,7 @@ module IRB
 	    print "Maybe IRB bug!!\n" if irb_bug
 	  end
           if $SAFE > 2
-            warn "Error: irb does not work for $SAFE level higher than 2"
-            exit 1
+            abort "Error: irb does not work for $SAFE level higher than 2"
           end
 	end
       end
@@ -304,7 +308,7 @@ module IRB
     def inspect
       ary = []
       for iv in instance_variables
-	case iv
+	case (iv = iv.to_s)
 	when "@signal_status"
 	  ary.push format("%s=:%s", iv, @signal_status.id2name)
 	when "@context"
