@@ -46,8 +46,16 @@
 require 'prettyprint'
 
 module Kernel
+  # returns a pretty printed object as a string.
+  def pretty_inspect
+    PP.pp(self, '')
+  end
+
   private
-  def pp(*objs)
+  # prints arguments in pretty form.
+  #
+  # pp returns nil.
+  def pp(*objs) # :doc:
     objs.each {|obj|
       PP.pp(obj)
     }
@@ -83,6 +91,12 @@ class PP < PrettyPrint
     out
   end
 
+  # :stopdoc:
+  def PP.mcall(obj, mod, meth, *args, &block)
+    mod.instance_method(meth).bind(obj).call(*args, &block)
+  end
+  # :startdoc:
+
   @sharing_detection = false
   class << self
     # Returns the sharing detection flag as a boolean value.
@@ -91,41 +105,55 @@ class PP < PrettyPrint
   end
 
   module PPMethods
-    InspectKey = :__inspect_key__
-
     def guard_inspect_key
-      if Thread.current[InspectKey] == nil
-        Thread.current[InspectKey] = []
+      if Thread.current[:__recursive_key__] == nil
+        Thread.current[:__recursive_key__] = {}
       end
 
-      save = Thread.current[InspectKey]
+      if Thread.current[:__recursive_key__][:inspect] == nil
+        Thread.current[:__recursive_key__][:inspect] = {}
+      end
+
+      save = Thread.current[:__recursive_key__][:inspect]
 
       begin
-        Thread.current[InspectKey] = []
+        Thread.current[:__recursive_key__][:inspect] = {}
         yield
       ensure
-        Thread.current[InspectKey] = save
+        Thread.current[:__recursive_key__][:inspect] = save
       end
     end
 
+    def check_inspect_key(id)
+      Thread.current[:__recursive_key__] &&
+      Thread.current[:__recursive_key__][:inspect] &&
+      Thread.current[:__recursive_key__][:inspect].include?(id)
+    end
+    def push_inspect_key(id)
+      Thread.current[:__recursive_key__][:inspect][id] = true
+    end
+    def pop_inspect_key(id)
+      Thread.current[:__recursive_key__][:inspect].delete id
+    end
+
     # Adds +obj+ to the pretty printing buffer
     # using Object#pretty_print or Object#pretty_print_cycle.
     # 
     # Object#pretty_print_cycle is used when +obj+ is already
     # printed, a.k.a the object reference chain has a cycle.
     def pp(obj)
-      id = obj.__id__
-
-      if Thread.current[InspectKey].include? id
+      id = obj.object_id
+ 
+      if check_inspect_key(id)
         group {obj.pretty_print_cycle self}
         return
       end
 
       begin
-        Thread.current[InspectKey] << id
+        push_inspect_key(id)
         group {obj.pretty_print self}
       ensure
-        Thread.current[InspectKey].pop unless PP.sharing_detection
+        pop_inspect_key(id) unless PP.sharing_detection
       end
     end
 
@@ -136,9 +164,23 @@ class PP < PrettyPrint
       group(1, '#<' + obj.class.name, '>', &block)
     end
 
+    if 0x100000000.class == Bignum
+      # 32bit
+      PointerMask = 0xffffffff
+    else
+      # 64bit
+      PointerMask = 0xffffffffffffffff
+    end
+
+    case Object.new.inspect
+    when /\A\#<Object:0x([0-9a-f]+)>\z/
+      PointerFormat = "%0#{$1.length}x"
+    else
+      PointerFormat = "%x"
+    end
+
     def object_address_group(obj, &block)
-      id = "%x" % (obj.__id__ * 2)
-      id.sub!(/\Af(?=[[:xdigit:]]{2}+\z)/, '') if id.sub!(/\A\.\./, '')
+      id = PointerFormat % (obj.object_id * 2 & PointerMask)
       group(1, "\#<#{obj.class}:0x#{id}", '>', &block)
     end
 
@@ -168,7 +210,7 @@ class PP < PrettyPrint
     # 
     #   q.seplist([1,2,3]) {|v| xxx v }
     # 
-    #   q.seplist([1,2,3], lambda { comma_breakable }, :each) {|v| xxx v }
+    #   q.seplist([1,2,3], lambda { q.comma_breakable }, :each) {|v| xxx v }
     # 
     #   xxx 1
     #   q.comma_breakable
@@ -241,9 +283,9 @@ class PP < PrettyPrint
     # This module provides predefined #pretty_print methods for some of
     # the most commonly used built-in classes for convenience.
     def pretty_print(q)
-      if /\(Kernel\)#/ !~ method(:inspect).inspect
+      if /\(Kernel\)#/ !~ Object.instance_method(:method).bind(self).call(:inspect).inspect
         q.text self.inspect
-      elsif /\(Kernel\)#/ !~ method(:to_s).inspect && instance_variables.empty?
+      elsif /\(Kernel\)#/ !~ Object.instance_method(:method).bind(self).call(:to_s).inspect && instance_variables.empty?
         q.text self.to_s
       else
         q.pp_object(self)
@@ -275,8 +317,8 @@ class PP < PrettyPrint
     # However, doing this requires that every class that #inspect is called on
     # implement #pretty_print, or a RuntimeError will be raised.
     def pretty_print_inspect
-      if /\(PP::ObjectMixin\)#/ =~ method(:pretty_print).inspect
-        raise "pretty_print is not overridden."
+      if /\(PP::ObjectMixin\)#/ =~ Object.instance_method(:method).bind(self).call(:pretty_print).inspect
+        raise "pretty_print is not overridden for #{self.class}"
       end
       PP.singleline_pp(self, '')
     end
@@ -309,14 +351,18 @@ end
 
 class << ENV
   def pretty_print(q)
-    q.pp_hash self
+    h = {}
+    ENV.keys.sort.each {|k|
+      h[k] = ENV[k]
+    }
+    q.pp_hash h
   end
 end
 
 class Struct
   def pretty_print(q)
-    q.group(1, '#<struct ' + self.class.name, '>') {
-      q.seplist(self.members, lambda { q.text "," }) {|member|
+    q.group(1, '#<struct ' + PP.mcall(self, Kernel, :class).name, '>') {
+      q.seplist(PP.mcall(self, Struct, :members), lambda { q.text "," }) {|member|
         q.breakable
         q.text member.to_s
         q.text '='
@@ -329,7 +375,7 @@ class Struct
   end
 
   def pretty_print_cycle(q)
-    q.text sprintf("#<struct %s:...>", self.class.name)
+    q.text sprintf("#<struct %s:...>", PP.mcall(self, Kernel, :class).name)
   end
 end
 
@@ -427,10 +473,24 @@ end
 
 class MatchData
   def pretty_print(q)
+    nc = []
+    self.regexp.named_captures.each {|name, indexes|
+      indexes.each {|i| nc[i] = name }
+    }
     q.object_group(self) {
       q.breakable
-      q.seplist(1..self.size, lambda { q.breakable }) {|i|
-        q.pp self[i-1]
+      q.seplist(0...self.size, lambda { q.breakable }) {|i|
+        if i == 0
+          q.pp self[i]
+        else
+          if nc[i]
+            q.text nc[i]
+          else
+            q.pp i
+          end
+          q.text ':'
+          q.pp self[i]
+        end
       }
     }
   end
@@ -468,6 +528,19 @@ if __FILE__ == $0
     def test_list0123_11
       assert_equal("[0,\n 1,\n 2,\n 3]\n", PP.pp([0,1,2,3], '', 11))
     end
+
+    OverriddenStruct = Struct.new("OverriddenStruct", :members, :class)
+    def test_struct_override_members # [ruby-core:7865]
+      a = OverriddenStruct.new(1,2)
+      assert_equal("#<struct Struct::OverriddenStruct members=1, class=2>\n", PP.pp(a, ''))
+    end
+
+    def test_redefined_method
+      o = ""
+      def o.method
+      end
+      assert_equal(%(""\n), PP.pp(o, ""))
+    end
   end
 
   class HasInspect
