@@ -28,6 +28,16 @@ static const char prelude_code0[] =
 "  end\n"
 "end\n"
 "\n"
+"def require_relative(relative_feature)\n"
+"  c = caller.first\n"
+"  e = c.rindex(/:\\d+:in /)\n"
+"  file = $`\n"
+"  if /\\A\\((.*)\\)/ =~ file # eval, etc.\n"
+"    raise LoadError, \"require_relative is called in #{$1}\"\n"
+"  end\n"
+"  absolute_feature = File.expand_path(File.join(File.dirname(file), relative_feature))\n"
+"  require absolute_feature\n"
+"end\n"
 ;
 
 void
