@@ -3,39 +3,42 @@
   range.c -
 
   $Author: nobu $
-  $Date: 2005/08/14 15:39:39 $
   created at: Thu Aug 19 17:46:47 JST 1993
 
-  Copyright (C) 1993-2003 Yukihiro Matsumoto
+  Copyright (C) 1993-2007 Yukihiro Matsumoto
 
 **********************************************************************/
 
-#include "ruby.h"
+#include "ruby/ruby.h"
+#include "ruby/encoding.h"
 
 VALUE rb_cRange;
 static ID id_cmp, id_succ, id_beg, id_end, id_excl;
 
-#define EXCL(r) RTEST(rb_ivar_get((r), id_excl))
-#define SET_EXCL(r,v) rb_ivar_set((r), id_excl, (v) ? Qtrue : Qfalse)
+extern VALUE rb_struct_init_copy(VALUE copy, VALUE s);
+
+#define RANGE_BEG(r) (RSTRUCT(r)->as.ary[0])
+#define RANGE_END(r) (RSTRUCT(r)->as.ary[1])
+#define RANGE_EXCL(r) (RSTRUCT(r)->as.ary[2])
+
+#define EXCL(r) RTEST(RANGE_EXCL(r))
+#define SET_EXCL(r,v) (RSTRUCT(r)->as.ary[2] = (v) ? Qtrue : Qfalse)
 
 static VALUE
-range_failed()
+range_failed(void)
 {
     rb_raise(rb_eArgError, "bad value for range");
     return Qnil;		/* dummy */
 }
 
 static VALUE
-range_check(args)
-    VALUE *args;
+range_check(VALUE *args)
 {
     return rb_funcall(args[0], id_cmp, 1, args[1]);
 }
 
 static void
-range_init(range, beg, end, exclude_end)
-    VALUE range, beg, end;
-    int exclude_end;
+range_init(VALUE range, VALUE beg, VALUE end, int exclude_end)
 {
     VALUE args[2];
 
@@ -46,18 +49,17 @@ range_init(range, beg, end, exclude_end)
 	VALUE v;
 
 	v = rb_rescue(range_check, (VALUE)args, range_failed, 0);
-	if (NIL_P(v)) range_failed();
+	if (NIL_P(v))
+	    range_failed();
     }
 
     SET_EXCL(range, exclude_end);
-    rb_ivar_set(range, id_beg, beg);
-    rb_ivar_set(range, id_end, end);
+    RSTRUCT(range)->as.ary[0] = beg;
+    RSTRUCT(range)->as.ary[1] = end;
 }
 
 VALUE
-rb_range_new(beg, end, exclude_end)
-    VALUE beg, end;
-    int exclude_end;
+rb_range_new(VALUE beg, VALUE end, int exclude_end)
 {
     VALUE range = rb_obj_alloc(rb_cRange);
 
@@ -75,22 +77,20 @@ rb_range_new(beg, end, exclude_end)
  */
 
 static VALUE
-range_initialize(argc, argv, range)
-    int argc;
-    VALUE *argv;
-    VALUE range;
+range_initialize(int argc, VALUE *argv, VALUE range)
 {
     VALUE beg, end, flags;
     
     rb_scan_args(argc, argv, "21", &beg, &end, &flags);
     /* Ranges are immutable, so that they should be initialized only once. */
-    if (rb_ivar_defined(range, id_beg)) {
+    if (RANGE_EXCL(range) != Qnil) {
 	rb_name_error(rb_intern("initialize"), "`initialize' called twice");
     }
     range_init(range, beg, end, RTEST(flags));
     return Qnil;
 }
 
+#define range_initialize_copy rb_struct_init_copy /* :nodoc: */
 
 /*
  *  call-seq:
@@ -100,8 +100,7 @@ range_initialize(argc, argv, range)
  */
 
 static VALUE
-range_exclude_end_p(range)
-    VALUE range;
+range_exclude_end_p(VALUE range)
 {
     return EXCL(range) ? Qtrue : Qfalse;
 }
@@ -122,45 +121,49 @@ range_exclude_end_p(range)
  */
 
 static VALUE
-range_eq(range, obj)
-    VALUE range, obj;
+range_eq(VALUE range, VALUE obj)
 {
-    if (range == obj) return Qtrue;
+    if (range == obj)
+	return Qtrue;
     if (!rb_obj_is_instance_of(obj, rb_obj_class(range)))
 	return Qfalse;
 
-    if (!rb_equal(rb_ivar_get(range, id_beg), rb_ivar_get(obj, id_beg)))
+    if (!rb_equal(RANGE_BEG(range), RANGE_BEG(obj)))
 	return Qfalse;
-    if (!rb_equal(rb_ivar_get(range, id_end), rb_ivar_get(obj, id_end)))
+    if (!rb_equal(RANGE_END(range), RANGE_END(obj)))
 	return Qfalse;
 
-    if (EXCL(range) != EXCL(obj)) return Qfalse;
+    if (EXCL(range) != EXCL(obj))
+	return Qfalse;
 
     return Qtrue;
 }
 
 static int
-r_lt(a, b)
-    VALUE a, b;
+r_lt(VALUE a, VALUE b)
 {
     VALUE r = rb_funcall(a, id_cmp, 1, b);
 
-    if (NIL_P(r)) return Qfalse;
-    if (rb_cmpint(r, a, b) < 0) return Qtrue;
+    if (NIL_P(r))
+	return Qfalse;
+    if (rb_cmpint(r, a, b) < 0)
+	return Qtrue;
     return Qfalse;
 }
 
 static int
-r_le(a, b)
-    VALUE a, b;
+r_le(VALUE a, VALUE b)
 {
     int c;
     VALUE r = rb_funcall(a, id_cmp, 1, b);
 
-    if (NIL_P(r)) return Qfalse;
+    if (NIL_P(r))
+	return Qfalse;
     c = rb_cmpint(r, a, b);
-    if (c == 0) return INT2FIX(0);
-    if (c < 0) return Qtrue;
+    if (c == 0)
+	return INT2FIX(0);
+    if (c < 0)
+	return Qtrue;
     return Qfalse;
 }
 
@@ -180,19 +183,20 @@ r_le(a, b)
  */
 
 static VALUE
-range_eql(range, obj)
-    VALUE range, obj;
+range_eql(VALUE range, VALUE obj)
 {
-    if (range == obj) return Qtrue;
+    if (range == obj)
+	return Qtrue;
     if (!rb_obj_is_instance_of(obj, rb_obj_class(range)))
 	return Qfalse;
 
-    if (!rb_eql(rb_ivar_get(range, id_beg), rb_ivar_get(obj, id_beg)))
+    if (!rb_eql(RANGE_BEG(range), RANGE_BEG(obj)))
 	return Qfalse;
-    if (!rb_eql(rb_ivar_get(range, id_end), rb_ivar_get(obj, id_end)))
+    if (!rb_eql(RANGE_END(range), RANGE_END(obj)))
 	return Qfalse;
 
-    if (EXCL(range) != EXCL(obj)) return Qfalse;
+    if (EXCL(range) != EXCL(obj))
+	return Qfalse;
 
     return Qtrue;
 }
@@ -207,59 +211,56 @@ range_eql(range, obj)
  */
 
 static VALUE
-range_hash(range)
-    VALUE range;
+range_hash(VALUE range)
 {
     long hash = EXCL(range);
     VALUE v;
 
-    v = rb_hash(rb_ivar_get(range, id_beg));
+    v = rb_hash(RANGE_BEG(range));
     hash ^= v << 1;
-    v = rb_hash(rb_ivar_get(range, id_end));
+    v = rb_hash(RANGE_END(range));
     hash ^= v << 9;
     hash ^= EXCL(range) << 24;
 
     return LONG2FIX(hash);
 }
 
-static VALUE
-str_step(args)
-    VALUE *args;
-{
-    return rb_str_upto(args[0], args[1], EXCL(args[2]));
-}
-
 static void
-range_each_func(range, func, v, e, arg)
-    VALUE range;
-    void (*func) _((VALUE, void*));
-    VALUE v, e;
-    void *arg;
+range_each_func(VALUE range, VALUE (*func) (VALUE, void *), void *arg)
 {
     int c;
+    VALUE b = RANGE_BEG(range);
+    VALUE e = RANGE_END(range);
+    VALUE v = b;
 
     if (EXCL(range)) {
 	while (r_lt(v, e)) {
-	    (*func)(v, arg);
+	    (*func) (v, arg);
 	    v = rb_funcall(v, id_succ, 0, 0);
 	}
     }
     else {
 	while (RTEST(c = r_le(v, e))) {
-	    (*func)(v, arg);
-	    if (c == INT2FIX(0)) break;
+	    (*func) (v, arg);
+	    if (c == INT2FIX(0))
+		break;
 	    v = rb_funcall(v, id_succ, 0, 0);
 	}
     }
 }
 
 static VALUE
-step_i(i, iter)
-    VALUE i;
-    long *iter;
+step_i(VALUE i, void *arg)
 {
-    iter[0]--;
-    if (iter[0] == 0) {
+    VALUE *iter = arg;
+
+    if (FIXNUM_P(iter[0])) {
+	iter[0] -= INT2FIX(1) & ~FIXNUM_FLAG;
+    }
+    else {
+	iter[0] = rb_funcall(iter[0], '-', 1, INT2FIX(1));
+    }
+    if (iter[0] == INT2FIX(0)) {
 	rb_yield(i);
 	iter[0] = iter[1];
     }
@@ -271,7 +272,7 @@ step_i(i, iter)
  *     rng.step(n=1) {| obj | block }    => rng
  *  
  *  Iterates over <i>rng</i>, passing each <i>n</i>th element to the block. If
- *  the range contains numbers or strings, natural ordering is used.  Otherwise
+ *  the range contains numbers, <i>n</i> is added for each iteration.  Otherwise
  *  <code>step</code> invokes <code>succ</code> to iterate through range
  *  elements. The following code uses class <code>Xs</code>, which is defined
  *  in the class-level documentation.
@@ -295,80 +296,87 @@ step_i(i, iter)
 
 
 static VALUE
-range_step(argc, argv, range)
-    int argc;
-    VALUE *argv;
-    VALUE range;
+range_step(int argc, VALUE *argv, VALUE range)
 {
-    VALUE b, e, step;
-    long unit;
+    VALUE b, e, step, tmp;
 
-    b = rb_ivar_get(range, id_beg);
-    e = rb_ivar_get(range, id_end);
-    if (rb_scan_args(argc, argv, "01", &step) == 0) {
+    RETURN_ENUMERATOR(range, argc, argv);
+
+    b = RANGE_BEG(range);
+    e = RANGE_END(range);
+    if (argc == 0) {
 	step = INT2FIX(1);
     }
+    else {
+	rb_scan_args(argc, argv, "01", &step);
+	if (!rb_obj_is_kind_of(step, rb_cNumeric)) {
+	    step = rb_to_int(step);
+	}
+	if (rb_funcall(step, '<', 1, INT2FIX(0))) {
+	    rb_raise(rb_eArgError, "step can't be negative");
+	}
+	else if (!rb_funcall(step, '>', 1, INT2FIX(0))) {
+	    rb_raise(rb_eArgError, "step can't be 0");
+	}
+    }
 
-    unit = NUM2LONG(step);
-    if (unit < 0) {
-	rb_raise(rb_eArgError, "step can't be negative");
-    } 
-    if (FIXNUM_P(b) && FIXNUM_P(e)) { /* fixnums are special */
+    if (FIXNUM_P(b) && FIXNUM_P(e) && FIXNUM_P(step)) { /* fixnums are special */
 	long end = FIX2LONG(e);
-	long i;
+	long i, unit = FIX2LONG(step);
 
-	if (unit == 0) rb_raise(rb_eArgError, "step can't be 0");
-	if (!EXCL(range)) end += 1;
-	for (i=FIX2LONG(b); i<end; i+=unit) {
+	if (!EXCL(range))
+	    end += 1;
+	i = FIX2LONG(b);	
+	while (i < end) {
 	    rb_yield(LONG2NUM(i));
+	    if (i + unit < i) break;
+	    i += unit;
+	}
+
+    }
+    else if (rb_obj_is_kind_of(b, rb_cNumeric) ||
+	     !NIL_P(rb_check_to_integer(b, "to_int")) ||
+	     !NIL_P(rb_check_to_integer(e, "to_int"))) {
+	ID op = EXCL(range) ? '<' : rb_intern("<=");
+
+	while (RTEST(rb_funcall(b, op, 1, e))) {
+	    rb_yield(b);
+	    b = rb_funcall(b, '+', 1, step);
 	}
     }
     else {
-	VALUE tmp = rb_check_string_type(b);
+	tmp = rb_check_string_type(b);
 
 	if (!NIL_P(tmp)) {
-	    VALUE args[5];
-	    long iter[2];
+	    VALUE args[2], iter[2];
 
 	    b = tmp;
-	    if (unit == 0) rb_raise(rb_eArgError, "step can't be 0");
-	    args[0] = b; args[1] = e; args[2] = range;
-	    iter[0] = 1; iter[1] = unit;
-	    rb_iterate((VALUE(*)_((VALUE)))str_step, (VALUE)args, step_i,
-			(VALUE)iter);
-	}
-	else if (rb_obj_is_kind_of(b, rb_cNumeric)) {
-	    ID c = rb_intern(EXCL(range) ? "<" : "<=");
-
-	    if (rb_equal(step, INT2FIX(0))) rb_raise(rb_eArgError, "step can't be 0");
-	    while (RTEST(rb_funcall(b, c, 1, e))) {
-		rb_yield(b);
-		b = rb_funcall(b, '+', 1, step);
-	    }
+	    args[0] = e;
+	    args[1] = EXCL(range) ? Qtrue : Qfalse;
+	    iter[0] = INT2FIX(1);
+	    iter[1] = step;
+	    rb_block_call(b, rb_intern("upto"), 2, args, step_i, (VALUE)iter);
 	}
 	else {
-	    long args[2];
+	    VALUE args[2];
 
-	    if (unit == 0) rb_raise(rb_eArgError, "step can't be 0");
 	    if (!rb_respond_to(b, id_succ)) {
 		rb_raise(rb_eTypeError, "can't iterate from %s",
 			 rb_obj_classname(b));
 	    }
-	
-	    args[0] = 1;
-	    args[1] = unit;
-	    range_each_func(range, step_i, b, e, args);
+	    args[0] = INT2FIX(1);
+	    args[1] = step;
+	    range_each_func(range, step_i, args);
 	}
     }
     return range;
 }
 
-static void
-each_i(v, arg)
-    VALUE v;
-    void *arg;
+static VALUE
+each_i(VALUE v, void *arg)
 {
     rb_yield(v);
+    return Qnil;
 }
 
 /*
@@ -390,13 +398,14 @@ each_i(v, arg)
  */
 
 static VALUE
-range_each(range)
-    VALUE range;
+range_each(VALUE range)
 {
     VALUE beg, end;
 
-    beg = rb_ivar_get(range, id_beg);
-    end = rb_ivar_get(range, id_end);
+    RETURN_ENUMERATOR(range, 0, 0);
+
+    beg = RANGE_BEG(range);
+    end = RANGE_END(range);
 
     if (!rb_respond_to(beg, id_succ)) {
 	rb_raise(rb_eTypeError, "can't iterate from %s",
@@ -406,46 +415,42 @@ range_each(range)
 	long lim = FIX2LONG(end);
 	long i;
 
-	if (!EXCL(range)) lim += 1;
-	for (i=FIX2LONG(beg); i<lim; i++) {
-	    rb_yield(LONG2NUM(i));
+	if (!EXCL(range))
+	    lim += 1;
+	for (i = FIX2LONG(beg); i < lim; i++) {
+	    rb_yield(LONG2FIX(i));
 	}
     }
     else if (TYPE(beg) == T_STRING) {
-	VALUE args[5];
-	long iter[2];
+	VALUE args[2];
 
-	args[0] = beg; args[1] = end; args[2] = range;
-	iter[0] = 1; iter[1] = 1;
-	rb_iterate((VALUE(*)_((VALUE)))str_step, (VALUE)args, step_i,
-		   (VALUE)iter);
+	args[0] = end;
+	args[1] = EXCL(range) ? Qtrue : Qfalse;
+	rb_block_call(beg, rb_intern("upto"), 2, args, rb_yield, 0);
     }
     else {
-	range_each_func(range, each_i, beg, end, NULL);
+	range_each_func(range, each_i, NULL);
     }
     return range;
 }
 
 /*
  *  call-seq:
- *     rng.first    => obj
  *     rng.begin    => obj
  *  
  *  Returns the first object in <i>rng</i>.
  */
 
 static VALUE
-range_first(range)
-    VALUE range;
+range_begin(VALUE range)
 {
-    return rb_ivar_get(range, id_beg);
+    return RANGE_BEG(range);
 }
 
 
 /*
  *  call-seq:
  *     rng.end    => obj
- *     rng.last   => obj
  *  
  *  Returns the object that defines the end of <i>rng</i>.
  *     
@@ -455,38 +460,174 @@ range_first(range)
 
 
 static VALUE
-range_last(range)
-    VALUE range;
+range_end(VALUE range)
 {
-    return rb_ivar_get(range, id_end);
+    return RANGE_END(range);
+}
+
+
+static VALUE
+first_i(VALUE i, VALUE *ary)
+{
+    long n = NUM2LONG(ary[0]);
+
+    if (n <= 0) {
+	rb_iter_break();
+    }
+    rb_ary_push(ary[1], i);
+    n--;
+    ary[0] = INT2NUM(n);
+    return Qnil;
+}
+
+/*
+ *  call-seq:
+ *     rng.first    => obj
+ *     rng.first(n) => an_array
+ *  
+ *  Returns the first object in <i>rng</i>, or the first +n+ elements.
+ */
+
+static VALUE
+range_first(int argc, VALUE *argv, VALUE range)
+{
+    VALUE n, ary[2];
+
+    if (argc == 0) return RANGE_BEG(range);
+
+    rb_scan_args(argc, argv, "1", &n);
+    ary[0] = n;
+    ary[1] = rb_ary_new2(NUM2LONG(n));
+    rb_block_call(range, rb_intern("each"), 0, 0, first_i, (VALUE)ary);
+
+    return ary[1];
+}
+
+
+/*
+ *  call-seq:
+ *     rng.last    => obj
+ *     rng.last(n) => an_array
+ *  
+ *  Returns the last object in <i>rng</i>, or the last +n+ elements.
+ */
+
+static VALUE
+range_last(int argc, VALUE *argv, VALUE range)
+{
+    VALUE rb_ary_last(int, VALUE *, VALUE);
+
+    if (argc == 0) return RANGE_END(range);
+    return rb_ary_last(argc, argv, rb_Array(range)); 
+}
+
+
+/*
+ *  call-seq:
+ *     rng.min                    => obj
+ *     rng.min {| a,b | block }   => obj
+ *  
+ *  Returns the minimum value in <i>rng</i>. The second uses
+ *  the block to compare values.  Returns nil if the first
+ *  value in range is larger than the last value.
+ *     
+ */
+
+
+static VALUE
+range_min(VALUE range)
+{
+    if (rb_block_given_p()) {
+	return rb_call_super(0, 0);
+    }
+    else {
+	VALUE b = RANGE_BEG(range);
+	VALUE e = RANGE_END(range);
+	int c = rb_cmpint(rb_funcall(b, id_cmp, 1, e), b, e);
+
+	if (c > 0 || (c == 0 && EXCL(range)))
+	    return Qnil;
+	return b;
+    }
+}
+
+/*
+ *  call-seq:
+ *     rng.max                    => obj
+ *     rng.max {| a,b | block }   => obj
+ *  
+ *  Returns the maximum value in <i>rng</i>. The second uses
+ *  the block to compare values.  Returns nil if the first
+ *  value in range is larger than the last value.
+ *     
+ */
+
+
+static VALUE
+range_max(VALUE range)
+{
+    VALUE e = RANGE_END(range);
+    int ip = FIXNUM_P(e) || rb_obj_is_kind_of(e, rb_cInteger);
+
+    if (rb_block_given_p() || (EXCL(range) && !ip)) {
+	return rb_call_super(0, 0);
+    }
+    else {
+	VALUE b = RANGE_BEG(range);
+	int c = rb_cmpint(rb_funcall(b, id_cmp, 1, e), b, e);
+
+	if (c > 0)
+	    return Qnil;
+	if (EXCL(range)) {
+	    if (c == 0) return Qnil;
+	    if (FIXNUM_P(e)) {
+		return LONG2NUM(FIX2LONG(e) - 1);
+	    }
+	    return rb_funcall(e, '-', 1, INT2FIX(1));
+	}
+	return e;
+    }
 }
 
 VALUE
-rb_range_beg_len(range, begp, lenp, len, err)
-    VALUE range;
-    long *begp, *lenp;
-    long len;
-    int err;
+rb_range_beg_len(VALUE range, long *begp, long *lenp, long len, int err)
 {
-    long beg, end, b, e;
+    VALUE b, e;
+    long beg, end, excl;
 
-    if (!rb_obj_is_kind_of(range, rb_cRange)) return Qfalse;
-
-    beg = b = NUM2LONG(rb_ivar_get(range, id_beg));
-    end = e = NUM2LONG(rb_ivar_get(range, id_end));
+    if (rb_obj_is_kind_of(range, rb_cRange)) {
+	b = RANGE_BEG(range);
+	e = RANGE_END(range);
+	excl = EXCL(range);
+    }
+    else {
+	if (!rb_respond_to(range, id_beg)) return Qfalse;
+	if (!rb_respond_to(range, id_end)) return Qfalse;
+	b = rb_funcall(range, id_beg, 0);
+	e = rb_funcall(range, id_end, 0);
+	excl = RTEST(rb_funcall(range, rb_intern("exclude_end?"), 0));
+    }
+    beg = NUM2LONG(b);
+    end = NUM2LONG(e);
 
     if (beg < 0) {
 	beg += len;
-	if (beg < 0) goto out_of_range;
+	if (beg < 0)
+	    goto out_of_range;
     }
     if (err == 0 || err == 2) {
-	if (beg > len) goto out_of_range;
-	if (end > len) end = len;
+	if (beg > len)
+	    goto out_of_range;
+	if (end > len)
+	    end = len;
     }
-    if (end < 0) end += len;
-    if (!EXCL(range)) end++;	/* include end point */
+    if (end < 0)
+	end += len;
+    if (!excl)
+	end++;			/* include end point */
     len = end - beg;
-    if (len < 0) len = 0;
+    if (len < 0)
+	len = 0;
 
     *begp = beg;
     *lenp = len;
@@ -495,7 +636,7 @@ rb_range_beg_len(range, begp, lenp, len, err)
   out_of_range:
     if (err) {
 	rb_raise(rb_eRangeError, "%ld..%s%ld out of range",
-		 b, EXCL(range)? "." : "", e);
+		 b, excl ? "." : "", e);
     }
     return Qnil;
 }
@@ -508,15 +649,14 @@ rb_range_beg_len(range, begp, lenp, len, err)
  */
 
 static VALUE
-range_to_s(range)
-    VALUE range;
+range_to_s(VALUE range)
 {
     VALUE str, str2;
 
-    str = rb_obj_as_string(rb_ivar_get(range, id_beg));
-    str2 = rb_obj_as_string(rb_ivar_get(range, id_end));
+    str = rb_obj_as_string(RANGE_BEG(range));
+    str2 = rb_obj_as_string(RANGE_END(range));
     str = rb_str_dup(str);
-    rb_str_cat(str, "...", EXCL(range)?3:2);
+    rb_str_cat(str, "...", EXCL(range) ? 3 : 2);
     rb_str_append(str, str2);
     OBJ_INFECT(str, str2);
 
@@ -534,15 +674,14 @@ range_to_s(range)
 
 
 static VALUE
-range_inspect(range)
-    VALUE range;
+range_inspect(VALUE range)
 {
     VALUE str, str2;
 
-    str = rb_inspect(rb_ivar_get(range, id_beg));
-    str2 = rb_inspect(rb_ivar_get(range, id_end));
+    str = rb_inspect(RANGE_BEG(range));
+    str2 = rb_inspect(RANGE_END(range));
     str = rb_str_dup(str);
-    rb_str_cat(str, "...", EXCL(range)?3:2);
+    rb_str_cat(str, "...", EXCL(range) ? 3 : 2);
     rb_str_append(str, str2);
     OBJ_INFECT(str, str2);
 
@@ -552,8 +691,6 @@ range_inspect(range)
 /*
  *  call-seq:
  *     rng === obj       =>  true or false
- *     rng.member?(val)  =>  true or false
- *     rng.include?(val) =>  true or false
  *  
  *  Returns <code>true</code> if <i>obj</i> is an element of
  *  <i>rng</i>, <code>false</code> otherwise. Conveniently,
@@ -572,24 +709,140 @@ range_inspect(range)
  */
 
 static VALUE
-range_include(range, val)
-    VALUE range, val;
+range_eqq(VALUE range, VALUE val)
+{
+    return rb_funcall(range, rb_intern("include?"), 1, val);
+}
+
+
+/*
+ *  call-seq:
+ *     rng.member?(val)  =>  true or false
+ *     rng.include?(val) =>  true or false
+ *  
+ *  Returns <code>true</code> if <i>obj</i> is an element of
+ *  <i>rng</i>, <code>false</code> otherwise.  If beg and end are
+ *  numeric, comparison is done according magnitude of values.
+ *     
+ *     ("a".."z").include?("g")  # => true
+ *     ("a".."z").include?("A")  # => false
+ */
+
+static VALUE
+range_include(VALUE range, VALUE val)
+{
+    VALUE beg = RANGE_BEG(range);
+    VALUE end = RANGE_END(range);
+    int nv = FIXNUM_P(beg) || FIXNUM_P(end) ||
+	     rb_obj_is_kind_of(beg, rb_cNumeric) ||
+	     rb_obj_is_kind_of(end, rb_cNumeric);
+
+    if (nv ||
+	!NIL_P(rb_check_to_integer(beg, "to_int")) ||
+	!NIL_P(rb_check_to_integer(end, "to_int"))) {
+	if (r_le(beg, val)) {
+	    if (EXCL(range)) {
+		if (r_lt(val, end))
+		    return Qtrue;
+	    }
+	    else {
+		if (r_le(val, end))
+		    return Qtrue;
+	    }
+	}
+	return Qfalse;
+    }
+    else if (TYPE(beg) == T_STRING && TYPE(end) == T_STRING &&
+	     RSTRING_LEN(beg) == 1 && RSTRING_LEN(end) == 1) {
+	if (NIL_P(val)) return Qfalse;
+	if (TYPE(val) == T_STRING) {
+	    if (RSTRING_LEN(val) == 0 || RSTRING_LEN(val) > 1)
+		return Qfalse;
+	    else {
+		char b = RSTRING_PTR(beg)[0];
+		char e = RSTRING_PTR(end)[0];
+		char v = RSTRING_PTR(val)[0];
+
+		if (ISASCII(b) && ISASCII(e) && ISASCII(v)) {
+		    if (b <= v && v < e) return Qtrue;
+		    if (!EXCL(range) && v == e) return Qtrue;
+		    return Qfalse;
+		}
+	    }
+	}
+    }
+    /* TODO: ruby_frame->this_func = rb_intern("include?"); */
+    return rb_call_super(1, &val);
+}
+
+
+/*
+ *  call-seq:
+ *     rng.cover?(val)  =>  true or false
+ *  
+ *  Returns <code>true</code> if <i>obj</i> is between beg and end,
+ *  i.e <code>beg <= obj <= end</code> (or <i>end</i> exclusive when
+ *  <code>exclude_end?</code> is true).
+ *     
+ *     ("a".."z").cover?("c")    #=> true
+ *     ("a".."z").cover?("5")    #=> false
+ */
+
+static VALUE
+range_cover(VALUE range, VALUE val)
 {
     VALUE beg, end;
 
-    beg = rb_ivar_get(range, id_beg);
-    end = rb_ivar_get(range, id_end);
+    beg = RANGE_BEG(range);
+    end = RANGE_END(range);
     if (r_le(beg, val)) {
 	if (EXCL(range)) {
-	    if (r_lt(val, end)) return Qtrue;
+	    if (r_lt(val, end))
+		return Qtrue;
 	}
 	else {
-	    if (r_le(val, end)) return Qtrue;
+	    if (r_le(val, end))
+		return Qtrue;
 	}
     }
     return Qfalse;
 }
 
+static VALUE
+range_dumper(VALUE range)
+{
+    VALUE v;
+    NEWOBJ(m, struct RObject);
+    OBJSETUP(m, rb_cObject, T_OBJECT);
+
+    v = (VALUE)m;
+
+    rb_ivar_set(v, id_excl, RANGE_EXCL(range));
+    rb_ivar_set(v, id_beg, RANGE_BEG(range));
+    rb_ivar_set(v, id_end, RANGE_END(range));
+    return v;
+}
+
+static VALUE
+range_loader(VALUE range, VALUE obj)
+{
+    if (TYPE(obj) != T_OBJECT || RBASIC(obj)->klass != rb_cObject) {
+        rb_raise(rb_eTypeError, "not a dumped range object");
+    }
+
+    RSTRUCT(range)->as.ary[0] = rb_ivar_get(obj, id_beg);
+    RSTRUCT(range)->as.ary[1] = rb_ivar_get(obj, id_end);
+    RSTRUCT(range)->as.ary[2] = rb_ivar_get(obj, id_excl);
+    return range;
+}
+
+static VALUE
+range_alloc(VALUE klass)
+{
+  /* rb_struct_alloc_noinit itself should not be used because
+   * rb_marshal_define_compat uses equality of allocaiton function */
+    return rb_struct_alloc_noinit(klass);
+}
 
 /*  A <code>Range</code> represents an interval---a set of values with a
  *  start and an end. Ranges may be constructed using the
@@ -644,21 +897,37 @@ range_include(range, val)
  */
 
 void
-Init_Range()
+Init_Range(void)
 {
-    rb_cRange = rb_define_class("Range", rb_cObject);
+#undef rb_intern
+#define rb_intern(str) rb_intern_const(str)
+
+    id_cmp = rb_intern("<=>");
+    id_succ = rb_intern("succ");
+    id_beg = rb_intern("begin");
+    id_end = rb_intern("end");
+    id_excl = rb_intern("excl");
+
+    rb_cRange = rb_struct_define_without_accessor(
+        "Range", rb_cObject, range_alloc,
+        "begin", "end", "excl", NULL);
+
     rb_include_module(rb_cRange, rb_mEnumerable);
+    rb_marshal_define_compat(rb_cRange, rb_cObject, range_dumper, range_loader);
     rb_define_method(rb_cRange, "initialize", range_initialize, -1);
+    rb_define_method(rb_cRange, "initialize_copy", range_initialize_copy, 1);
     rb_define_method(rb_cRange, "==", range_eq, 1);
-    rb_define_method(rb_cRange, "===", range_include, 1);
+    rb_define_method(rb_cRange, "===", range_eqq, 1);
     rb_define_method(rb_cRange, "eql?", range_eql, 1);
     rb_define_method(rb_cRange, "hash", range_hash, 0);
     rb_define_method(rb_cRange, "each", range_each, 0);
     rb_define_method(rb_cRange, "step", range_step, -1);
-    rb_define_method(rb_cRange, "first", range_first, 0);
-    rb_define_method(rb_cRange, "last", range_last, 0);
-    rb_define_method(rb_cRange, "begin", range_first, 0);
-    rb_define_method(rb_cRange, "end", range_last, 0);
+    rb_define_method(rb_cRange, "begin", range_begin, 0);
+    rb_define_method(rb_cRange, "end", range_end, 0);
+    rb_define_method(rb_cRange, "first", range_first, -1);
+    rb_define_method(rb_cRange, "last", range_last, -1);
+    rb_define_method(rb_cRange, "min", range_min, 0);
+    rb_define_method(rb_cRange, "max", range_max, 0);
     rb_define_method(rb_cRange, "to_s", range_to_s, 0);
     rb_define_method(rb_cRange, "inspect", range_inspect, 0);
 
@@ -666,10 +935,5 @@ Init_Range()
 
     rb_define_method(rb_cRange, "member?", range_include, 1);
     rb_define_method(rb_cRange, "include?", range_include, 1);
-
-    id_cmp = rb_intern("<=>");
-    id_succ = rb_intern("succ");
-    id_beg = rb_intern("begin");
-    id_end = rb_intern("end");
-    id_excl = rb_intern("excl");
+    rb_define_method(rb_cRange, "cover?", range_cover, 1);
 }
