@@ -1,4 +1,5 @@
 require 'test/unit'
+require_relative 'envutil'
 
 class TestVariable < Test::Unit::TestCase
   class Gods
@@ -25,8 +26,11 @@ class TestVariable < Test::Unit::TestCase
   end
 
   class Titans < Gods
-    @@rule = "Cronus"
-    include Olympians           	# OK to cause warning (intentional)
+    @@rule = "Cronus"			# modifies @@rule in Gods
+    include Olympians
+    def ruler4
+      @@rule
+    end
   end
 
   def test_variable
@@ -49,5 +53,31 @@ class TestVariable < Test::Unit::TestCase
     atlas = Titans.new
     assert_equal("Cronus", atlas.ruler0)
     assert_equal("Zeus", atlas.ruler3)
+    assert_equal("Cronus", atlas.ruler4)
+  end
+
+  def test_local_variables
+    lvar = 1
+    assert_instance_of(Symbol, local_variables[0], "[ruby-dev:34008]")
+  end
+
+  def test_local_variables2
+    x = 1
+    proc do |y|
+      assert_equal([:x, :y], local_variables.sort)
+    end.call
+  end
+
+  def test_local_variables3
+    x = 1
+    proc do |y|
+      1.times do |z|
+        assert_equal([:x, :y, :z], local_variables.sort)
+      end
+    end.call
+  end
+
+  def test_global_variable_0
+    assert_in_out_err(["-e", "$0='t'*1000;print $0"], "", /\At+\z/, [])
   end
 end
